@@ -3,9 +3,9 @@
  *
  * TOMOYO Linux's utilities.
  *
- * Copyright (C) 2005-2011  NTT DATA CORPORATION
+ * Copyright (C) 2005-2012  NTT DATA CORPORATION
  *
- * Version: 2.4.0+   2011/09/29
+ * Version: 2.5.0+   2012/03/13
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License v2 as published by the
@@ -955,6 +955,258 @@ int ccs_parse_number(const char *number, struct ccs_number_entry *entry)
 	return 0;
 }
 
+/*
+ * Routines for parsing IPv4 or IPv6 address.
+ * These are copied from lib/hexdump.c net/core/utils.c .
+ */
+#include <ctype.h>
+
+static int hex_to_bin(char ch)
+{
+	if ((ch >= '0') && (ch <= '9'))
+		return ch - '0';
+	ch = tolower(ch);
+	if ((ch >= 'a') && (ch <= 'f'))
+		return ch - 'a' + 10;
+	return -1;
+}
+
+#define IN6PTON_XDIGIT		0x00010000
+#define IN6PTON_DIGIT		0x00020000
+#define IN6PTON_COLON_MASK	0x00700000
+#define IN6PTON_COLON_1		0x00100000	/* single : requested */
+#define IN6PTON_COLON_2		0x00200000	/* second : requested */
+#define IN6PTON_COLON_1_2	0x00400000	/* :: requested */
+#define IN6PTON_DOT		0x00800000	/* . */
+#define IN6PTON_DELIM		0x10000000
+#define IN6PTON_NULL		0x20000000	/* first/tail */
+#define IN6PTON_UNKNOWN		0x40000000
+
+static inline int xdigit2bin(char c, int delim)
+{
+	int val;
+
+	if (c == delim || c == '\0')
+		return IN6PTON_DELIM;
+	if (c == ':')
+		return IN6PTON_COLON_MASK;
+	if (c == '.')
+		return IN6PTON_DOT;
+
+	val = hex_to_bin(c);
+	if (val >= 0)
+		return val | IN6PTON_XDIGIT | (val < 10 ? IN6PTON_DIGIT : 0);
+
+	if (delim == -1)
+		return IN6PTON_DELIM;
+	return IN6PTON_UNKNOWN;
+}
+
+static int in4_pton(const char *src, int srclen, u8 *dst, int delim,
+		    const char **end)
+{
+	const char *s;
+	u8 *d;
+	u8 dbuf[4];
+	int ret = 0;
+	int i;
+	int w = 0;
+
+	if (srclen < 0)
+		srclen = strlen(src);
+	s = src;
+	d = dbuf;
+	i = 0;
+	while(1) {
+		int c;
+		c = xdigit2bin(srclen > 0 ? *s : '\0', delim);
+		if (!(c & (IN6PTON_DIGIT | IN6PTON_DOT | IN6PTON_DELIM |
+			   IN6PTON_COLON_MASK))) {
+			goto out;
+		}
+		if (c & (IN6PTON_DOT | IN6PTON_DELIM | IN6PTON_COLON_MASK)) {
+			if (w == 0)
+				goto out;
+			*d++ = w & 0xff;
+			w = 0;
+			i++;
+			if (c & (IN6PTON_DELIM | IN6PTON_COLON_MASK)) {
+				if (i != 4)
+					goto out;
+				break;
+			}
+			goto cont;
+		}
+		w = (w * 10) + c;
+		if ((w & 0xffff) > 255) {
+			goto out;
+		}
+cont:
+		if (i >= 4)
+			goto out;
+		s++;
+		srclen--;
+	}
+	ret = 1;
+	memcpy(dst, dbuf, sizeof(dbuf));
+out:
+	if (end)
+		*end = s;
+	return ret;
+}
+
+static int in6_pton(const char *src, int srclen, u8 *dst, int delim,
+		    const char **end)
+{
+	const char *s, *tok = NULL;
+	u8 *d, *dc = NULL;
+	u8 dbuf[16];
+	int ret = 0;
+	int i;
+	int state = IN6PTON_COLON_1_2 | IN6PTON_XDIGIT | IN6PTON_NULL;
+	int w = 0;
+
+	memset(dbuf, 0, sizeof(dbuf));
+
+	s = src;
+	d = dbuf;
+	if (srclen < 0)
+		srclen = strlen(src);
+
+	while (1) {
+		int c;
+
+		c = xdigit2bin(srclen > 0 ? *s : '\0', delim);
+		if (!(c & state))
+			goto out;
+		if (c & (IN6PTON_DELIM | IN6PTON_COLON_MASK)) {
+			/* process one 16-bit word */
+			if (!(state & IN6PTON_NULL)) {
+				*d++ = (w >> 8) & 0xff;
+				*d++ = w & 0xff;
+			}
+			w = 0;
+			if (c & IN6PTON_DELIM) {
+				/* We've processed last word */
+				break;
+			}
+			/*
+			 * COLON_1 => XDIGIT
+			 * COLON_2 => XDIGIT|DELIM
+			 * COLON_1_2 => COLON_2
+			 */
+			switch (state & IN6PTON_COLON_MASK) {
+			case IN6PTON_COLON_2:
+				dc = d;
+				state = IN6PTON_XDIGIT | IN6PTON_DELIM;
+				if (dc - dbuf >= sizeof(dbuf))
+					state |= IN6PTON_NULL;
+				break;
+			case IN6PTON_COLON_1|IN6PTON_COLON_1_2:
+				state = IN6PTON_XDIGIT | IN6PTON_COLON_2;
+				break;
+			case IN6PTON_COLON_1:
+				state = IN6PTON_XDIGIT;
+				break;
+			case IN6PTON_COLON_1_2:
+				state = IN6PTON_COLON_2;
+				break;
+			default:
+				state = 0;
+			}
+			tok = s + 1;
+			goto cont;
+		}
+
+		if (c & IN6PTON_DOT) {
+			ret = in4_pton(tok ? tok : s, srclen + (int)(s - tok),
+				       d, delim, &s);
+			if (ret > 0) {
+				d += 4;
+				break;
+			}
+			goto out;
+		}
+
+		w = (w << 4) | (0xff & c);
+		state = IN6PTON_COLON_1 | IN6PTON_DELIM;
+		if (!(w & 0xf000)) {
+			state |= IN6PTON_XDIGIT;
+		}
+		if (!dc && d + 2 < dbuf + sizeof(dbuf)) {
+			state |= IN6PTON_COLON_1_2;
+			state &= ~IN6PTON_DELIM;
+		}
+		if (d + 2 >= dbuf + sizeof(dbuf)) {
+			state &= ~(IN6PTON_COLON_1|IN6PTON_COLON_1_2);
+		}
+cont:
+		if ((dc && d + 4 < dbuf + sizeof(dbuf)) ||
+		    d + 4 == dbuf + sizeof(dbuf)) {
+			state |= IN6PTON_DOT;
+		}
+		if (d >= dbuf + sizeof(dbuf)) {
+			state &= ~(IN6PTON_XDIGIT|IN6PTON_COLON_MASK);
+		}
+		s++;
+		srclen--;
+	}
+
+	i = 15; d--;
+
+	if (dc) {
+		while(d >= dc)
+			dst[i--] = *d--;
+		while(i >= dc - dbuf)
+			dst[i--] = 0;
+		while(i >= 0)
+			dst[i--] = *d--;
+	} else
+		memcpy(dst, dbuf, sizeof(dbuf));
+
+	ret = 1;
+out:
+	if (end)
+		*end = s;
+	return ret;
+}
+
+/**
+ * ccs_parse_ip - Parse a ccs_ip_address_entry.
+ *
+ * @address: IP address or IP range.
+ * @entry:   Pointer to "struct ccs_address_entry".
+ *
+ * Returns 0 on success, -EINVAL otherwise.
+ */
+int ccs_parse_ip(const char *address, struct ccs_ip_address_entry *entry)
+{
+	u8 * const min = entry->min;
+	u8 * const max = entry->max;
+	const char *end;
+	memset(entry, 0, sizeof(*entry));
+	if (!strchr(address, ':') &&
+	    in4_pton(address, -1, min, '-', &end) > 0) {
+		entry->is_ipv6 = false;
+		if (!*end)
+			memmove(max, min, 4);
+		else if (*end++ != '-' ||
+			 in4_pton(end, -1, max, '\0', &end) <= 0 || *end)
+			return -EINVAL;
+		return 0;
+	}
+	if (in6_pton(address, -1, min, '-', &end) > 0) {
+		entry->is_ipv6 = true;
+		if (!*end)
+			memmove(max, min, 16);
+		else if (*end++ != '-' ||
+			 in6_pton(end, -1, max, '\0', &end) <= 0 || *end)
+			return -EINVAL;
+		return 0;
+	}
+	return -EINVAL;
+}
+
 /**
  * ccs_open_stream - Establish IP connection.
  *
@@ -985,24 +1237,18 @@ int ccs_open_stream(const char *filename)
  *
  * @dp:          Pointer to "const struct ccs_domain_policy".
  * @domainname0: Name of domain to find.
- * @is_dis:      True if the domain acts as domain initialize source, false
- *               otherwise.
- * @is_dd:       True if the domain is marked as deleted, false otherwise.
  *
  * Returns index number (>= 0) in the @dp array if found, EOF otherwise.
  */
 int ccs_find_domain(const struct ccs_domain_policy *dp,
-		    const char *domainname0, const _Bool is_dis,
-		    const _Bool is_dd)
+		    const char *domainname0)
 {
 	int i;
 	struct ccs_path_info domainname;
 	domainname.name = domainname0;
 	ccs_fill_path_info(&domainname);
 	for (i = 0; i < dp->list_len; i++) {
-		if (dp->list[i].is_dis == is_dis &&
-		    dp->list[i].is_dd == is_dd &&
-		    !ccs_pathcmp(&domainname, dp->list[i].domainname))
+		if (!ccs_pathcmp(&domainname, dp->list[i].domainname))
 			return i;
 	}
 	return EOF;
@@ -1013,20 +1259,16 @@ int ccs_find_domain(const struct ccs_domain_policy *dp,
  *
  * @dp:         Pointer to "struct ccs_domain_policy".
  * @domainname: Name of domain to find.
- * @is_dis:     True if the domain acts as domain initialize source, false
- *              otherwise.
- * @is_dd:      True if the domain is marked as deleted, false otherwise.
  *
  * Returns index number (>= 0) in the @dp array if created or already exists,
  * abort otherwise.
  */
-int ccs_assign_domain(struct ccs_domain_policy *dp, const char *domainname,
-		      const _Bool is_dis, const _Bool is_dd)
+int ccs_assign_domain(struct ccs_domain_policy *dp, const char *domainname)
 {
-	int index = ccs_find_domain(dp, domainname, is_dis, is_dd);
+	int index = ccs_find_domain(dp, domainname);
 	if (index >= 0)
 		return index;
-	if (!is_dis && !ccs_correct_domain(domainname)) {
+	if (!ccs_correct_domain(domainname)) {
 		fprintf(stderr, "Invalid domainname '%s'\n", domainname);
 		ccs_out_of_memory();
 	}
@@ -1035,8 +1277,6 @@ int ccs_assign_domain(struct ccs_domain_policy *dp, const char *domainname,
 			       sizeof(struct ccs_domain_info));
 	memset(&dp->list[index], 0, sizeof(struct ccs_domain_info));
 	dp->list[index].domainname = ccs_savename(domainname);
-	dp->list[index].is_dis = is_dis;
-	dp->list[index].is_dd = is_dd;
 	return index;
 }
 
@@ -1731,19 +1971,17 @@ void ccs_handle_domain_policy(struct ccs_domain_policy *dp, FILE *fp,
 		ccs_str_starts(line, "domain=");
 		if (ccs_domain_def(line)) {
 			if (is_delete) {
-				index = ccs_find_domain(dp, line, false,
-							false);
+				index = ccs_find_domain(dp, line);
 				if (index >= 0)
 					ccs_delete_domain(dp, index);
 				index = EOF;
 				continue;
 			}
 			if (is_select) {
-				index = ccs_find_domain(dp, line, false,
-							false);
+				index = ccs_find_domain(dp, line);
 				continue;
 			}
-			index = ccs_assign_domain(dp, line, false, false);
+			index = ccs_assign_domain(dp, line);
 			continue;
 		}
 		if (index == EOF || !line[0])
@@ -1926,7 +2164,7 @@ char *ccs_freadline_unpack(FILE *fp)
 		else
 			offset = 0;
 		/*
-		 * Only "file " is subjected to unpacking.
+		 * Only "file " and "network " are subjected to unpacking.
 		 */
 		if (!strncmp(line + offset, "file ", 5)) {
 			char *cp = line + offset + 5;
@@ -1936,6 +2174,19 @@ char *ccs_freadline_unpack(FILE *fp)
 				pack_start = cp - line;
 				goto prepare;
 			}
+		} else if (!strncmp(line + offset, "network ", 8)) {
+			char *cp = strchr(line + offset + 8, ' ');
+			char *cp2 = NULL;
+			if (cp)
+				cp = strchr(cp + 1, ' ');
+			if (cp)
+				cp2 = strchr(cp + 1, ' ');
+			cp++;
+			len = cp2 - cp;
+			if (cp2 && memchr(cp, '/', len)) {
+				pack_start = cp - line;
+				goto prepare;
+			}
 		}
 		return line;
 prepare:
@@ -1979,9 +2230,9 @@ unpack:
 }
 
 /**
- * ccs_check_remote_host - Check whether the remote host is running with the TOMOYO 2.4 kernel or not.
+ * ccs_check_remote_host - Check whether the remote host is running with the TOMOYO 2.5 kernel or not.
  *
- * Returns true if running with TOMOYO 2.4 kernel, false otherwise.
+ * Returns true if running with TOMOYO 2.5 kernel, false otherwise.
  */
 _Bool ccs_check_remote_host(void)
 {
@@ -1991,7 +2242,7 @@ _Bool ccs_check_remote_host(void)
 	FILE *fp = ccs_open_read("version");
 	if (!fp ||
 	    fscanf(fp, "%u.%u.%u", &major, &minor, &rev) < 2 ||
-	    major != 2 || minor != 4) {
+	    major != 2 || minor != 5) {
 		const u32 ip = ntohl(ccs_network_ip);
 		fprintf(stderr, "Can't connect to %u.%u.%u.%u:%u\n",
 			(u8) (ip >> 24), (u8) (ip >> 16),
